#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .LayerClass import Layer

class PropagationLayer(Layer):
    """Handle all the traced layer non define as call_module or call_function/method.

    :param node: Is the actual traced node from torch.fx.
    :type node: torch.fx.node.Node
    :param trace: Is the complete trace of torch.fx of the model.
    :type trace: torch.fx.graph._node_list
    :param prev_node: Is the just previous node in the trace before the actual traced node.
    :type node: torch.fx.node.Node
    """

    def __init__(self, node, trace, prev_node):
        """Constructor."""
        super().__init__(node, trace, prev_node)

    def get_declaration(self) -> str:
        """Generate and return the full class generated for a propagation layer.
        
        What we call a propagation layer, is a layer who do no direct action on the data but propagate it further.

        Sometimes this layer have a special argument to propagate but have to return the input for the next layers.

        :return: The full declaration of a Layer containing a CallFunction
        :rtype: str
        """
        string = self.generate_class()

        task = ""
        if len(self.node.args) > 0 and self.node.op == "placeholder":
            task = str(self.node.args[0])
        else:
            task = "input"

        string += self.generate_forward(task)

        return string

    def __str__(self) -> str:
        """Allow to print easily all the information of a layer.
        
        It adds a print to inform that it is a propagation layer.
        :return: String to print
        :rtype: str
        """
        print_str = super().__str__()
        print_str += "    This layer is just a propagation one\n\n"
        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch
from ..function_parser import _parse_func

class GetAttr:
    """Handle all the getattr.
    
    A getattr cannot be a layer because the result of a getattr is not a tensor.
    So we have to attach the getattr with the Layer who need to use a getattr on his attribute.
    For that we create an object who save all the information, parent, child and attribute for the getattr.
    :param node: The actual getattr node
    :type node: torch.fx.node.Node
    :param trace: The trace of the model generated by torch fx
    :type trace: torch.fx.graph._node_list
    """

    def __init__(self, node, trace, child = None):
        """Constructor."""
        self.getitem_idx = None
        self.parent = node.args[0]
        self.attr = node.args[1]
        self.stash_needed = False
        self.node = node
        self.position = 0
        
        self.child = child
        if self.child is None:
            for _node in trace:
                if node in _node.args:
                    if str(_node).find("getitem") >= 0:
                        node = _node
                    else:
                        self.child = _node
                        self.position = _node.args.index(node)
                        break
        # Verifiy if the result of the getattr need to be stashed.
        prev_node = None
        for _node in trace:
            if _node == self.child:
                if prev_node != self.parent:
                    self.stash_needed = True
                    self.getattr_string = f"{self.parent}.{self.attr}"
                else:
                    self.getattr_string = f"input.{self.attr}"

            if not str(_node).find("getitem") >= 0 and self.node != _node:
                prev_node = _node
    
    def get_position(self) -> int:
        """Return the position of the attribute in the list of param during the call
        
        :return: Return the position of the attribute in the list of param during the call
        :rtype: int
        """
        return self.position

    def get_attr_name(self) -> str:
        """Return the name of the attribute used.

        :return: Return the name of the attribute used
        :rtype: str
        """
        return self.getattr_string

    def get_child(self) -> torch.fx.node.Node:
        """Return the node in which the getattr will be done.

        :return: Return the child node of the getattr
        :rtype: torch.fx.node.Node
        """
        return self.child

    def get_parent(self) -> torch.fx.node.Node:
        """Return the node of the ret value on which will be done the getattr.

        :return: Return the parent node of the getattr
        :rtype: torch.fx.node.Node
        """
        if self.stash_needed:
            return self.parent
        else:
            return 'input'

    def add_getitem(self, idx):
        """Allow to add a getitem on a getattr call.
        
        For example : input.shape[0]

        :param idx: The index of the getitem
        :type idx: int
        """
        self.getitem_idx = idx
        self.getattr_string += f"[{self.getitem_idx}]"

    def is_stash_needed(self) -> bool:
        """Return True if the getattr need to be stashed, else False.
        
        :return: If stash is needed
        :rtype: bool
        """
        return self.stash_needed

    def __str__(self) -> str:
        """Allow to print easily all the information of a Getattr.

        :return: String to print
        :rtype: str
        """
        print_str = "GetAttr info : \n"
        print_str += f"    Attribute {self.attr} of ret of node {self.parent}\n"
        print_str += f"    This getattr is needed in {self.child}\n"
        print_str += f"    If not empty use those specific idx of the result {self.getitem_idx}\n"
        if self.stash_needed:
            print_str += f"    A new stash have to be add on {self.parent} for {self.child}\n"
        else:
            print_str += f"    No new stash need to be added on {self.parent}, the dest is directly connected\n"

        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from gpu_alloc import TraceMalloc
from dataset import PipelineDataset
from pipelinecache.layered_model import PipelinedModel
import os
import time
import torch
import argparse
os.environ['MASTER_ADDR'] = 'localhost'
os.environ['MASTER_PORT'] = '29600'

parser = argparse.ArgumentParser()
parser.add_argument('--input_shape', type=str, help='Input shape as a list')
parser.add_argument('--output_shape', type=str, help='Output shape as a list')
parser.add_argument('--number_gpu', type=int, help='Number of GPU')
parser.add_argument('--number_chunks', type=int, help='Number of chunks')
parser.add_argument('--dtype', type=str,  help='Type of output\'s tensor (long, float32...)')
args = parser.parse_args()

input_shape = args.input_shape.replace("[", "").replace("]", "")
input_shape = input_shape.split(",")
input_shape = [int(x.strip()) for x in input_shape]

output_shape  = args.output_shape.replace("[", "").replace("]", "")
output_shape  = output_shape.split(",")
output_shape  = [int(x.strip()) for x in output_shape]

number_gpus   = args.number_gpu
number_chunks = args.number_chunks


trace_gpu_alloc = TraceMalloc(number_gpus)
criterion = torch.nn.CrossEntropyLoss()

torch.cuda.init()
torch.distributed.rpc.init_rpc('worker', rank=0, world_size=1)

with trace_gpu_alloc:

    model = PipelinedModel()
    dataset = PipelineDataset(1024, input_shape[1:], [1] if len(output_shape) == 1 else output_shape[1:], args.dtype)
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=input_shape[0], shuffle=True)

    model = model.get_modules()
    model = torch.distributed.pipeline.sync.Pipe(model, number_chunks)
    optimizer = torch.optim.SGD(model.parameters(), lr=0.001)

    for i in range(3):
        start_time = time.time()
        for inputs, labels in dataloader:
            optimizer.zero_grad()
            inputs = inputs.to(0)
            labels = labels.to(number_gpus- 1)

            try:
                outputs = model(inputs).local_value()
            except Exception as e:
                print(e)
                exit()

            labels_tmp = labels.squeeze()
            
            loss = criterion(outputs, labels.squeeze())
                
            loss.backward()
        end_time = time.time()
        execution_time = end_time - start_time

print(trace_gpu_alloc.peaked)





#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch
from ..constant import TAB

class Layer:
    """Set up of the Layer to manage every thing for creating a layer with the init and forward functions.

    This class is the base class for CallFunction, CallModule, GetAttrModule and PropagationLayer. Thus,
    this should never have called.

    :param node: Is the actual traced node from torch.fx.
    :type node: torch.fx.node.Node
    :param trace: Is the complete trace of torch.fx of the model.
    :type trace: torch.fx.graph._node_list
    :param prev_node: Is the just previous node in the trace before the actual traced node.
    :type node: torch.fx.node.Node
    """

    def __init__(self, node, trace, prev_node):
        """Constructor."""
        self.node = node
        self.name = str(node)
        self.args = []
        self.kwargs = []
        self.pop_list = []
        self.stash_list = []

        self.getitem_idx = None
        # True if there is a change of CPU at this specific layer
        self.separation_layer = False

        # Call function to handle correctly all information about the node.
        self._args_handler(trace, prev_node)
        self._kwargs_handler(trace, prev_node)
        self._pop_handler(prev_node)

    def get_node(self) -> torch.fx.node.Node:
        """Getter for the node treated.

        :return: Node of the layer
        :rtype: torch.fx.node.Node
        """
        return self.node

    def get_name(self) -> str:
        """Getter for the name of the node, just is string format.

        :return: Name of the node.
        :rtype: str
        """
        return self.name

    def set_stash(self, target_node):
        """Update the stash_list, i.e. the list of layers to which it must send the result of its forward.

        :param target_node: Node for which the result should be saved.
        :type target_node: torch.fx.node.Node
        """
        tmp = self.name + "_to_" + str(target_node)
        if tmp not in self.stash_list:
            self.stash_list.append(tmp)

    def _get_handler(self, arg, trace) -> torch.fx.node.Node:
        """Exist in the only objective to handler successive getattr and getitem call.
        
        Because those traced node should not be considered as Layer, and we have to search for the last node before them.
        :param arg: The arg to check
        :type arg: torch.fx.node.Node
        :param trace: Is the complete trace of torch.fx of the model.
        :type trace: torch.fx.graph._node_list
        :return: The last parent of the chained getitem and getattr call (if there is no chain just their direct parent)
        :rtype: torch.fx.node.Node
        """
        _arg = arg
        while str(_arg).find("getitem") >= 0 or str(_arg).find("getattr") >= 0:
            for node in trace:
                if node == _arg:
                    _arg = node.args[0]
                    break
        return _arg

    def update_arg_by_attr(self, new_arg, position):
        """Allow to update argument with a specific attribute.
        
        For example t.shape
        :param new_arg: New arg containing the attribute. (For ex .shape)
        :type new_arg: str
        :param position: Position of the argument in the arg list
        :type position: int
        """
        self.args[position] = new_arg


    def get_pop_parent(self) -> list:
        """Getter for the pop_list, the list of the layer who the actual layers need stashed value.

        :return: The pop list
        :rtype: list
        """
        return self.pop_list

    def _pop_handler(self, prev_node):
        """Handle the creation of the pop list.
        
        This function will go throw the args of the treated node and found if in his param he have a layer who is not directly connected to him.
        :param prev_node: The node directly before the treated node
        :type prev_node: torch.fx.node.Node
        """
        for arg in self.args:
            if isinstance(arg, tuple):
                tmp = arg[1]
            else:
                tmp = arg
            # If the node is not instance of torch.fx, then it's a param numeric or other..
            if isinstance(tmp, torch.fx.node.Node) and tmp != prev_node:
                # ocmon avoid to have multiple time the same pop in the pop list. Because we can have several time
                # the same value in the params.
                ocmon = False
                for pop in self.pop_list:
                    if str(pop[0]) == str(tmp):
                        ocmon = True
                        break
                if not ocmon:
                    self.pop_list.append([tmp, str(tmp) + '_to_' + self.name])

    def _kwargs_handler(self, trace, prev_node):
        """Handle the kwargs.
        
        The kwargs are specific and need to be writen as "dim=1". 
        In this this function save the name of the parameter and his value in the list of args.
        """
        for key, kwarg in self.node.kwargs.items():
            kwarg = self._get_handler(kwarg, trace)

            if kwarg == prev_node:
                self.args.append((key, "input"))
            else:
                self.args.append((key, kwarg))
            

    def _args_handler(self, trace, prev_node):
        """Handle the args.
        
        Args can be values but also the returned value of a past layer, so we have tocheck if the args is older than the direct previous or not. 
        Also if in the args there is a getattr or getitem we have to delete it with a call to _get_handler()
        :param trace: Is the complete trace of torch.fx of the model.
        :type trace: torch.fx.graph._node_list
        :param prev_node: The node directly before the treated node
        :type prev_node: torch.fx.node.Node
        """
        def process_argument(obj, arg, trace):
            if isinstance(arg, list):
                for item in arg:
                    process_argument(obj, item, trace)
            else:
                arg = self._get_handler(arg, trace)
                if arg == prev_node:
                    obj.args.append("input")
                else:
                    obj.args.append(arg)

        for arg in self.node.args:
            process_argument(self, arg, trace)

    def generate_class(self) -> str:
        """Create a string containing the full class declaration for a layer.

        For example if there is any pop or stash to make : 
        .. code-block:: python
            @skippable(stash=[..], pop=[..])
        And then the class :
        .. code-block:: python
            class self.name_layer(nn.Module)

        :return: Return the class declaration
        :rtype: str
        """
        string = ""
        if len(self.stash_list) > 0 or len(self.pop_list) > 0:
            string += f"@skippable("

            if len(self.stash_list) > 0:
                string += "stash=["
                for stash in self.stash_list:
                    string += f"'{stash}', "
                string = string[:-2]
                string += "], "

            if len(self.pop_list) > 0:
                string += "pop=["
                for pop in self.pop_list:
                    string += f"'{pop[1]}', "
                string = string[:-2]
                string += "], "

            string = string[:-2]
            string += ")\n"

        string += f"class {self.name}_layer(nn.Module):\n"
        return string

    def generate_init(self, declaration) -> str:
        """Create a string containing the full __init__ function of the layer.

        For example :
        .. code-block:: python
            def __init__(self) -> None:
                super().__init__()
                self.fc = nn.DECLARATION

        The declaration can be for example nn.Linear(in_features=4, out_features=16, bias=True) or nn.parameter.Parameter(1, 16, requires_grad=True)

        :param declaration: Is the module or parameter to initialize
        :type declaration: str
        :return: Return the class __init__
        :rtype: str
        """
        string = TAB[1] + "def __init__(self) -> None:\n"
        string += TAB[2] + "super().__init__()\n"
        string += TAB[2] + f"self.fc = nn.{declaration}\n"

        return string

    def reset_separation_layer(self):
        self.separation_layer = False

    def set_separation_layer(self):
        """Set the layer as a separation layer.
        
        This is needed because when the data change from a GPU to another he needs to be cloned.

        So this adds to the forward declaration : input = input.clone()
        """
        self.separation_layer = True

    def get_separation_layer(self) -> bool:
        """Return the separation status of the layer.
        
        :return: True if the layer is a separation layer, else False
        :rtype: bool
        """
        return self.separation_layer

    def generate_forward(self, task) -> str:
        """Create the full forward definition.
        
        He has to first, if needed, pop all the params sent from older layer. 
        Secondly clone the input if it is a separation layer. 
        Thirdly do his task and affect it to the ret variable. A task can be some different thing, for example the execution of the Module declarated in the __init__ (self.fc) or simply a function call (torch.add).
        Fourthly it will add, if defined, a getitem to the executed task.  So if the task return Tuple and only the first one is used it will add at the end a [0]. 
        Fifthly he have to stash is ret value if anyone need it further.
        And finaly return is value for the next layer.

        A full forward can look as :
        .. code-block:: python
            def forward(self, input):
                add = yield pop('add_to_{self.name}')
                input = input.clone()
                ret = torch.add(input, add)
                yield stash('{self.name}_to_another', ret)
                return ret

        :param task: Task to be done in the forward
        :type task: str
        :return: Return the full forward declaration
        :rtype: str
        """
        string = TAB[1] + "def forward(self, input):\n"

        # # TO REMOVE !!!!!
        # string += TAB[2] + f"with open(\"f.txt\", \"a\") as f:\n"
        # string += TAB[3] + f"f.write(f\"{self.node} - \")\n"
        # # ---------------

        for pop in self.pop_list:
            string += TAB[2] + f"{pop[0]} = yield pop('{pop[1]}')\n"

        if self.separation_layer:
            string += TAB[2] + "input = input.clone()\n"

        string += TAB[2] + f"ret = {task}"

        if self.getitem_idx is not None:
            string += f"[{self.getitem_idx}]"

        string += "\n"

        for stash in self.stash_list:
            string += TAB[2] + f"yield stash('{stash}', ret)\n"

        if self.node.op == "placeholder":
            string += TAB[2] + f"return input\n\n"
        else:
            string += TAB[2] + f"return ret\n\n"

        return string

    def add_stash(self, node_to):
        """Add stash information for special layer.
        
        The getattr are ignore during the parsing, so if one of those depend on layer we have to add their parent to the stash list to not lose the connection.
        :param node_to: Parent of a getattr
        :type node_to: torch.fx.node.Node
        """
        self.stash_list.append(self.name + "_to_" + str(node_to))

    def add_getitem(self, idx):
        """Allow to add a getitem to the task of the layer.
        
        Because the getitem cannot be considered as layer, we have to link the getitem to the layer with who he is linked.
        :param idx: Index of the getitem
        :type idx: int
        """
        self.getitem_idx = idx

    def __str__(self) -> str:
        """Allow to print easily all the information of a layer.

        :return: String to print
        :rtype: str
        """
        print_str = "---- Layer information ----\n"
        print_str += f"    Name : {self.name}\n"
        print_str += f"    Argument : {self.args}\n"

        if len(self.pop_list) > 0:
            print_str += f"    Pop info :\n"

            for pop, stashed_from in self.pop_list:
                print_str += f"        Argument {pop}, Stashed from {stashed_from}\n"

        if len(self.stash_list) > 0:
            print_str += "    Stash info : \n"

            for stash in self.stash_list:
                print_str += f"        ret is stashed for {stash}\n"
        if self.getitem_idx is not None:
            print_str += f"    This node will access the result of ret with [{self.getitem_idx}]\n"
        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

def _generate_function(var_names, func) -> str:
    """Return a string containg a call to a torch function.
    
    :param var_names: All the var that will be passed to the torch function
    :type var_names: list
    :param func: Name of the torch function to call
    :type func: str
    """
    string = f"torch.{func}("
    for var in var_names:        
        if isinstance(var, tuple):
            string += f"{var[0]}={var[1]}, "
        
        else:
            if isinstance(var, list):
                string += f"["
                for elem in var:
                    string += f"{elem}, "
                string = string[:-2]
                string += f"], "
            else:
                string += f"{var}, "
    string = string[:-2]
    string += ")"
    return string

def _generate_method(var_names, meth) -> str:
    """Return a string containg a call to a tensor method.
    
    :param var_names: All the var that will be passed to the tensor method
    :type var_names: list
    :param func: Name of the tensor method
    :type func: str
    """
    string = f"{var_names[0]}.{meth}("
    if len(var_names) > 1:
        for var in var_names[1:]:
            if isinstance(var, tuple):
                string += f"{var[0]}={var[1]}, "
            else:
                if isinstance(var, list):
                    string += f"["
                    for elem in var:
                        string += f"{elem}, "
                    string = string[:-2]
                    string += f"], "
                else:
                    string += f"{var}, "
        string = string[:-2]
    string += ")"
    return string

def gen_var_names(var_names, scheme, has_kwargs) -> list: 
    """Return the modify arg list with the specified scheme.

    :param var_names: List of argument.
    :type var_names: list of list, tuple, torch.fx.Node.node, int and string
    :param scheme: Description of argument repartition by group.
    :type scheme: list of int
    :param has_kwargs: Number of kwargs in var_names
    :type has_kwargs: int
    :return: List of grouped var_names with the description of scheme
    :rtype: list of list, tuple, torch.fx.Node.node, int and string
    """
    arg_cnt = 0
    new_var_names = []
    for n in scheme:
        if n > 1:
            new_var_names.append(var_names[arg_cnt:arg_cnt+n])
        else:
            new_var_names.append(var_names[arg_cnt])
        arg_cnt += n
    if has_kwargs:
        new_var_names.extend(var_names[-has_kwargs:])

    return new_var_names

def _parse_func(key, var_names) -> str:
    """Return a string containg the full call to a method or function for torch tensor.
    
    :param key: Node that have his op call_method or call_function
    :type key: torch.fx.Node.node
    :param var_names: All the var that will be passed to the torch function
    :type var_names: list
    """
    name = str(key)
    try:
        if name.find("add") >= 0:  
            return _generate_function(var_names, "add")

        elif name.find("sub") >= 0:  
            return _generate_function(var_names, "sub")

        elif name.find("mul") >= 0 and name.find("matmul") < 0:  
            return _generate_function(var_names, "mul")

        elif name.find("floordiv") >= 0 or name.find("floor_divide") >= 0:   
            return _generate_function(var_names, "floor_divide")

        elif name.find("truediv") >= 0 or name.find("true_divide") >= 0:  
            return _generate_function(var_names, "true_divide")

        elif name.find("cat") >= 0:
            tmp_var = var_names
            has_kwargs = 0

            while isinstance(tmp_var[-1], tuple):
                tmp_var = tmp_var[:-1]
                has_kwargs += 1

            if isinstance(tmp_var[-1], int):
                scheme = [len(tmp_var[:-1]), 1]
            else:
                scheme = [len(tmp_var)]

            var_names = gen_var_names(var_names, scheme, has_kwargs)

            return _generate_function(var_names, "cat")

        elif name.find("split") >= 0:  
            tmp_var = var_names
            has_kwargs = 0

            while isinstance(tmp_var[-1], tuple):
                tmp_var = tmp_var[:-1]
                has_kwargs += 1

            scheme = [1, len(tmp_var[1:])]
            
            var_names = gen_var_names(var_names, scheme, has_kwargs)

            return _generate_function(var_names, "split")

        elif name.find("flatten") >= 0:  
            return _generate_function(var_names, "flatten")

        elif name.find("relu") >= 0:  
            return _generate_function(var_names, "nn.functional.relu")

        elif name.find("matmul") >= 0:  
            return _generate_function(var_names, "matmul") 

        elif name.find("transpose") >= 0:  
            return _generate_function(var_names, "transpose")

        elif name.find("expand") >= 0:  
            return _generate_method(var_names, "expand")

        elif name.find("reshape") >= 0: 
            tmp_var = var_names

            has_kwargs = 0

            while isinstance(tmp_var[-1], tuple):
                tmp_var = tmp_var[:-1]
                has_kwargs += 1
            
            scheme = [1, len(tmp_var[1:])]
            
            var_names = gen_var_names(var_names, scheme, 0)

            return _generate_function(var_names, "reshape")

        elif name.find("permute") >= 0: 
            tmp_var = var_names
            has_kwargs = 0

            while isinstance(tmp_var[-1], tuple):
                tmp_var = tmp_var[:-1]
                has_kwargs += 1
            
            scheme = [1, len(tmp_var[1:])]
            
            var_names = gen_var_names(var_names, scheme, 0)

            return _generate_function(var_names, "permute")

        elif name.find("softmax") >= 0:  
            return _generate_function(var_names, "nn.functional.softmax")

        elif name.find("view") >= 0:  
            return _generate_method(var_names, "view")

        elif name.find("to") >= 0:  
            return _generate_method(var_names, "to")

        elif name.find("pow") >= 0:  
            return _generate_function(var_names, "pow")

        elif name.find("mean") >= 0:   
            return _generate_function(var_names, "mean")

        elif name.find("rsqrt") >= 0:  
            return _generate_function(var_names, "rsqrt")

        elif name.find("unsqueeze") >= 0:  
            return _generate_function(var_names, "unsqueeze")

        elif name.find("squeeze") >= 0:  
            return _generate_method(var_names, "squeeze")

        elif name.find("float") >= 0: 
            raise AssertionError(f"Function not handled: {name}")

        elif name.find("type_as") >= 0: 
            raise AssertionError(f"Function not handled: {name}")

        elif name.find("dropout") >= 0:  
            return _generate_function(var_names, "nn.functional.dropout")

        elif name.find("contiguous") >= 0: 
            raise AssertionError(f"Function not handled: {name}")

        elif name.find("tanh") >= 0:  
            return _generate_function(var_names, "nn.functional.tanh")

        elif name.find("gelu") >= 0:  
            return _generate_function(var_names, "nn.functional.gelu")

        elif name.find("size") >= 0:  
            return _generate_method(var_names, "size")

        else:
            raise AssertionError(f"Unknown function or method: {name}")
        
    except AssertionError as e:
        print(f"Error : {e}") 
        exit()


#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .LayerClass import Layer
from ..function_parser import _parse_func


class CallFunction(Layer):
    """Handle all the traced layer define as "call_function".

    A call function is a layer with no init, he just executes a function or method on tensors.
    For example a torch.add or a call to the method expand.

    :param node: Is the actual traced node from torch.fx.
    :type node: torch.fx.node.Node
    :param trace: Is the complete trace of torch.fx of the model.
    :type trace: torch.fx.graph._node_list
    :param prev_node: Is the just previous node in the trace before the actual traced node.
    :type node: torch.fx.node.Node
    """
    
    def __init__(self, node, trace, prev_node):
        """Constructor."""
        super().__init__(node, trace, prev_node)
        # This call will return the exact function or method call
        self.function_call = _parse_func(node, self.args)

    def update_arg_by_attr(self, new_arg, position):
        """Allow to update argument with a specific attribute.

        For example t.shape. We need to reset the function call as we change the args.

        :param old_arg: Old argument to replace.
        :type old_arg: torch.fx.node.Node
        :param new_arg: New arg containing the attribute. (For ex .shape)
        :type new_arg: str
        """
        super().update_arg_by_attr(new_arg, position)
        self.function_call = _parse_func(self.node, self.args)

    def get_declaration(self) -> str:
        """Generate and return the full class generate for a layer containing a call method or call function.
        
        For example :
        .. code-block:: python
            @skippable...
            class {self.name}_layer(nn.Module):
                def forward(self, input):
                    ... = yield pop...
                    ret = call function or call method (input.expand(16, -1, -1))
                    yield stash ...
                    return ret

        :return: The full declaration of a Layer containing a CallFunction
        :rtype: str
        """
        string = self.generate_class()
        string += self.generate_forward(self.function_call)

        return string

    def __str__(self) -> str:
        """Allow to print easily all the information of a layer.
        
        It adds a print of the function executed at the overload of the Layer class.

        :return: String to print
        :rtype: str
        """
        print_str = super().__str__()
        print_str += f"    The function call is {self.function_call}\n\n"
        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch

class PipelineDataset(torch.utils.data.Dataset):
    def __init__(self, size, input_shape, output_shape, dtype="long"):
        """Initialize a PipelineDataset object.
        
        :param size: Size of the dataset.
        :type size: int
        :param input_shape: Shape of the input data.
        :type input_shape: tuple
        :param output_shape: Shape of the output data.
        :type output_shape: tuple
        """
        self.size = size
        self.input_shape = input_shape
        self.output_shape = output_shape
        self.dtype = self.parse_dtype(dtype)

    def parse_dtype(self, dtype):
        """Map data type in string format to return data type of API torch

        :param dtype: Data type
        :type dtype: str
        :return: Data type of API torch
        :rtype: dtype
        """
        data_type_mapping = {
            'float32': torch.float32,
            'float': torch.float32,
            'float64': torch.float64,
            'double': torch.float64,
            'float16': torch.float16,
            'half': torch.float16,
            'bfloat16': torch.bfloat16,
            'complex32': torch.complex32,
            'chalf': torch.complex32,
            'complex64': torch.complex64,
            'cfloat': torch.complex64,
            'complex128': torch.complex128,
            'cdouble': torch.complex128,
            'uint8': torch.uint8,
            'byte': torch.uint8,
            'int8': torch.int8,
            'char': torch.int8,
            'int16': torch.int16,
            'short': torch.int16,
            'int32': torch.int32,
            'int': torch.int32,
            'int64': torch.int64,
            'long': torch.int64,
            'bool': torch.bool,
            'quint8': torch.quint8,
            'qint8': torch.qint8,
            'qint32': torch.qint32,
            'quint4x2': torch.quint4x2
        }
        data_type_str = dtype.lower()  # Convertir en minuscules pour être insensible à la casse
        return data_type_mapping.get(dtype, None)

    def __len__(self):
        """Return the length of the dataset.
        
        :return: Length of the dataset.
        :rtype: int
        """
        return self.size

    def __getitem__(self, idx):
        """Get an item from the dataset at the given index.
        
        :param idx: Index of the item.
        :type idx: int
        :return: Tuple of input data and target data.
        :rtype: tuple
        """
        data = torch.randn(*self.input_shape)
        target = torch.randint(0, 2, self.output_shape, dtype=self.dtype)
        return data, target




#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .LayerClass import Layer
import torch


class GetAttrModule(Layer):
    """Handle all the traced layer define as "getattr".
    
    If a special self is used in the module traced, torch fx mark it as an getattr layer.
    So we have to recreate the self object used and propagate it.

    For the moment only nn.parameter.Parameter are handled.

    :param node: Is the actual traced node from torch.fx.
    :type node: torch.fx.node.Node
    :param trace: Is the complete trace of torch.fx of the model.
    :type trace: torch.fx.graph._node_list
    :param prev_node: Is the just previous node in the trace before the actual traced node.
    :type node: torch.fx.node.Node
    :param net: The network given to the pipeline tool.
    :type net: Non-specific, could be a module or a set of module.
    """

    def __init__(self, node, trace, prev_node, net):
        """Constructor."""
        super().__init__(node, trace, prev_node)
        # We have to split the node.target because he contains the name of the attribute used.
        # For example "model.pooling.layer.query" the name of the attribute is the last word, query.
        module_tmp = str(node.target).split('.')
        tmp = ""
        for name in module_tmp[0:-1]:
            tmp += f"{name}."
        tmp = tmp[:-1]
        attr_parsed = module_tmp[len(module_tmp) - 1:]

        # Once we have the attribute to found we will search in the network given by the user
        for name, module in net.named_modules():
            if str(name) == tmp:
                attr = getattr(module, attr_parsed[0])

                if isinstance(attr, torch.nn.parameter.Parameter):
                    self.module_attr_desc = "parameter.Parameter(torch.Tensor("

                    for shape in attr.shape:
                        self.module_attr_desc += f"{shape}, "

                    self.module_attr_desc = self.module_attr_desc[:-2]

                    requires_grad = getattr(attr, "requires_grad")

                    self.module_attr_desc += f"), requires_grad={requires_grad})"

    def get_declaration(self) -> str:
        """Generate and return the full class generate for a layer containing a gettatr of a module.
        
        For example :
        .. code-block:: python
            @skippable...
            class {self.name}_layer(nn.Module):
                def __init__(self) -> None:
                    super().__init__()
                    self.fc = nn.parameter.Parameter(torch.Tensor(1, 16), requires_grad=True)
                def forward(self, input):
                    ret = self.fc
                    yield stash ...
                    return ret
        :return: The full class generated for a getattr layer
        :rtype: str
        """
        string = self.generate_class()
        string += self.generate_init(str(self.module_attr_desc))
        string += self.generate_forward("self.fc")

        return string

    def __str__(self) -> str:
        """Allow to print easily all the information of a layer.
        
        It adds a print of the parameter created in the layer.
        :return: String to print
        :rtype: str
        """
        print_str = super().__str__()
        print_str += f"    The parameter description is {self.module_attr_desc}\n\n"
        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

import torch

def BToMb(x):
    """Convert bytes to megabytes.
    
    :param x: Value in bytes.
    :type x: int
    :return: Value in megabytes.
    :rtype: int
    """
    return (x // (2 * 1024))

class TraceMalloc():
    def __init__(self, nb_gpu):
        """Initialize a TraceMalloc object.
        
        :param nb_gpu: Number of GPUs.
        :type nb_gpu: int
        """
        self.nb_gpu = nb_gpu
        self.begin  = [0] * nb_gpu
        self.end    = [0] * nb_gpu
        self.peak   = [0] * nb_gpu
        self.peaked = [0] * nb_gpu

    def __enter__(self):
        """Enter the context manager. 
        
        Save the current memory allocated to all GPUs.
        
        :return: The TraceMalloc object.
        :rtype: TraceMalloc
        """
        for device in range(self.nb_gpu):
            self.begin[device] = torch.cuda.memory_allocated(device)
            
        return self
    
    def __exit__(self, *exc):
        """Exit the context manager. 

        Get all the memory information, allocated and peak, to calculate the true peak between the enter and exit call.
        """
        for device in range(self.nb_gpu):
            self.end[device]    = torch.cuda.memory_allocated(device)
            self.peak[device]   = torch.cuda.max_memory_allocated(device)
            self.peaked[device] = BToMb(self.peak[device] - self.begin[device])
            torch.cuda.reset_peak_memory_stats(device)


#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .LayerClass import Layer

class CallModule(Layer):
    """Handle all the traced layer define as "call_module".
    
    Modules are all the torch Module present in the file "whitelist.txt", we do not guaranty the handle of custom module.

    :param node: Is the actual traced node from torch.fx.
    :type node: torch.fx.node.Node
    :param trace: Is the complete trace of torch.fx of the model.
    :type trace: torch.fx.graph._node_list
    :param prev_node: Is the just previous node in the trace before the actual traced node.
    :type node: torch.fx.node.Node
    """

    def __init__(self, node, trace, prev_node, module_desc):
        """Constructor."""
        super().__init__(node, trace, prev_node)
        self.module_desc = module_desc

    def get_declaration(self) -> str:
        """Generate and return the full class generate for a layer containing a call method or call function.
        
        For example :
        .. code-block:: python
            @skippable...
            class {self.name}_layer(nn.Module):
                def __init__(self) -> None:
                    super().__init__()
                    self.fc = nn.call_module
                def forward(self, input):
                    ... = yield pop...
                    ret = call module (self.fc(input))
                    yield stash ...
                    return ret

        :return: The full declaration of a Layer containing a CallModule
        :rtype: str
        """
        string = self.generate_class()
        string += self.generate_init(str(self.module_desc))

        task = f"self.fc("
        for arg in self.args:
            # This is for handle the kwargs (i.e. dim=1)
            if isinstance(arg, tuple):
                task += f"{arg[0]}={arg[1]}, "
            else:
                task += f"{arg}, "
        task = task[:-2]
        task += ")"

        string += self.generate_forward(task)

        return string

    def __str__(self) -> str:
        """Allow to print easily all the information of a layer.
        
        It adds a print of the torch module executed.
        :return: String to print
        :rtype: str
        """
        print_str = super().__str__()
        print_str += f"    The module description is {self.module_desc}\n\n"
        return print_str



#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

class PipelineConfig:
    """Create a configuration object for pipeline tool.

    :param input_shape: Input shape of the model
    :type input_shape: list of int
    :param output_shape: Output shape of the model
    :type output_shape: list of int
    :param data_type: Data type given to the model
    :type data_type: str
    :param config_mha: Multihead Attention configuration, allow reconstituion
    :type config_mha: dict
    """
    def __init__(self, input_shape, output_shape, data_type, config_mha = []) -> None:
        """Constructor."""
        self.input_shape = input_shape
        self.output_shape = output_shape
        self.data_type = data_type
        self.config_mha = config_mha

    def create_mha_conf_equal(self, nb_mha, num_heads, embed_dim, dropout, batch_first):
        """Allow to easily create Multihead configuration if all are equals.

        :param nb_mha: Number of multihead in the model
        :type nb_mha: int
        :param num_heads: Number of parallel attention heads.
        :type num_heads: int
        :param embed_dim: Total dimension of the Multihead.
        :type embed_dim: int
        :param dropout: Dropout probability
        :type dropout: float
        :param batch_first: Define if yes or not the batch is first in the input and output tensor.
        :type batch_first: bool
        """
        self.config_mha = []
        for i in range(nb_mha):
            self.config_mha.append({'embed_dim': embed_dim, 'num_heads': num_heads, 'dropout': dropout, 'batch_first': batch_first})


    @classmethod
    def orbit5k(cls):
        config_mha = [{'embed_dim': 16, 'num_heads': 8, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 16, 'num_heads': 8, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 16, 'num_heads': 8, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 16, 'num_heads': 8, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 16, 'num_heads': 8, 'dropout': 0.1, 'batch_first': True}]
        return cls([4, 1290, 4], [4], "long", config_mha)
    
    @classmethod
    def orbit5kbig(cls):
        config_mha = [{'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True},
                      {'embed_dim': 128, 'num_heads': 32, 'dropout': 0.1, 'batch_first': True}]
        return cls([4, 1309, 4], [4], "long", config_mha)
    
    def video_transform(cls):
        config_mha = [{'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}, {'embed_dim': 1280, 'num_heads': 16, 'dropout': 0.0, 'batch_first': True}]
        return cls([4, 3, 518, 518], [4, 1000], "float", config_mha)

    


#    The Pipeline tool allows you to create a new model, split across multiple GPUs, 
#      from a PyTorch module. This enables the training of large models that do not fit on a single GPU.
#    Copyright (C) 2023  Bruno Da Rocha Carvalho, Gabriel Catel Torres Arzur
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

from torch.fx import symbolic_trace
import torch
from pathlib import Path
from .constant import TAB
import subprocess
import sys
from .class_impl.CallModule import CallModule
from .class_impl.CallFunction import CallFunction
from .class_impl.PropagationLayer import PropagationLayer
from .class_impl.GetAttr import GetAttr
from .class_impl.GetAttrModule import GetAttrModule
import logging

class SkippableTracing:
    """Create and sequence the model parallelism.

    He will parsed the model given by the user and generate a file that contain all the splitted Layer and divide them on the gpus.

    Example of use :
    .. code-block:: python
        trace = SkippableTracing(2, model)
        # Then just get the generated splitted model
        model = trace.get_modules()

    :param nb_gpus: Nb of gpus of work, if none given max gpus will be taken.
    :type nb_gpus: int
    :param model: Model to parsed
    :type model: a model extended of nn.Module.
    """
    def __init__(self, nb_gpus, model, config):
        """Constructor."""
        self.module_desc = {}
        self.file = ""
        self.net = None
        self.LayerLists = {}
        self.GetattrLists = {}
        self.nb_gpu = nb_gpus if nb_gpus is not None else torch.cuda.device_count()
        self.file_name = "layered_model.py"
        self.directory_name = "pipelinecache"
        self.configs_mha = config.config_mha
        self.mha_number = len(self.configs_mha) 
        self.mha_count = 0
        self.input_shape = config.input_shape
        self.output_shape = config.output_shape
        self.dtype = config.data_type

        # logging.basicConfig(filename='pipelinecache/debug.log', encoding='utf-8', level=logging.DEBUG)
        
        self._verfiy_config_mha()

        self._tracer(model)

    def _verfiy_config_mha(self):
        """Verify if at least embed_dim and num_heads are present in the configuration given by the user.
        
        This two parameters are mandatory to create a MHA.
        """
        for config in self.configs_mha:
            try:
                config['embed_dim']
            except KeyError as e:
                raise KeyError("You didn't provide embed_dim in one of your MHA config")
            
            try:
                config['num_heads']
            except KeyError as e:
                raise KeyError("You didn't provide num_heads in one of your MHA config")

    def _write_in_file(self):
        """Write to the output file the generated Layers to use it from other files."""        
        dir_path = Path(__file__).resolve().parent / self.directory_name

        if not dir_path.exists():
            dir_path.mkdir(parents=True)

        file_path = dir_path / self.file_name

        with open(file_path, "w") as f:
            f.write(self.file)
            f.close()

    def get_modules(self) -> torch.nn.Sequential:
        """Allow the user to get the generated Sequential model for each GPU."""
        from .pipelinecache.layered_model import PipelinedModel

        model = PipelinedModel()
        return model.get_modules()

    def _init_file(self):
        """Add all necessary import to the file."""
        self.file += "import torch\n"
        self.file += "import torch.nn.functional as F\n"
        self.file += "import torch.nn as nn\n"
        self.file += "from torch.distributed.pipeline.sync.skip import stash, pop, skippable \n\n"

    def _generate_end_class(self):
        """Add a class at the end of the generated file to get simply the pipelined model."""
        self.file += f"class PipelinedModel(nn.Module):\n"
        self.file += TAB[1] + "def __init__(self) -> None:\n"
        self.file += TAB[2] + "super().__init__()\n"

        gpu_index = 0
        self.file += TAB[2] + f"self.s{gpu_index} = nn.Sequential("

        for layer in self.LayerLists.values():
            self.file += f"{layer.get_name()}_layer(), "

            if layer.get_separation_layer():
                self.file = self.file[:-2]
                self.file += f").cuda({gpu_index})\n"
                gpu_index = gpu_index + 1
                self.file += TAB[2] + f"self.s{gpu_index} = nn.Sequential("

        self.file = self.file[:-2]
        self.file += f").cuda({gpu_index})\n"

        self.file += TAB[1] + "def forward(self, input):\n"
        self.file += TAB[2] + f"ret = input\n"
        for gpu in range(self.nb_gpu):
            self.file += TAB[2] + f"ret = self.s{gpu}(ret.to({gpu}))\n"

        self.file += TAB[2] + "return ret\n"

        self.file += TAB[1] + "def get_modules(self):\n"
        self.file += TAB[2] + "return  nn.Sequential(*["
        for gpu in range(self.nb_gpu):
            self.file += f"nn.Sequential(*self.s{gpu}),"
        self.file = self.file[:-1]
        self.file += "])\n"

    def _create_mha(self):
        """Create MHA string declaration."""
        config = self.configs_mha[self.mha_count]

        decl = f"MultiheadAttention("

        for key, param in config.items():
            decl += f"{key}={param}, "

        decl = decl[:-2]
        decl += ")"
        self.mha_count = self.mha_count + 1

        return decl

    def _catch_module_desc(self):
        """Create a look-up dictionary to match target names with their declaration.
        
        We use the withelist.txt to know which module name we have to keep as "core" module.
        All the modules not present in the whitelist willbe digged to found their composition.

        MultiheadAttention are not parsed, so we have to do a little trick to found how is it configured, only based on giotto_deep implementation.
        """
        filename = Path(__file__).resolve().parent / 'whitelist.txt'
        whitelist = open(filename).readlines()
        whitelist = [line.strip() for line in whitelist]

        for name, module in self.net.named_modules():
            if str(module).split('(', 1)[0] in whitelist:
                if str(module).split('(', 1)[0].find('Multi') >= 0:
                    # try:
                    if self.mha_number >= 1:
                        self.module_desc[name] = self._create_mha()
                    else:
                        raise UserWarning(f"Error: You didn't specified any MHA config, but at least one exist.")
                    # except UserWarning as e:
                    #     raise Exception(f"Error : {e}")
                else:
                    self.module_desc[name] = module

    def _balancing(self, layers, memory):
        """Balance the distribution of layers across GPUs based on memory usage.
    
        :param layers: Current distribution of layers across GPUs.
        :type layers: list
        :param memory: Memory usage for each GPU.
        :type memory: list
        :return: New balanced distribution of layers across GPUs.
        :rtype: list
        """
        repartition = layers.copy()
        memory_tmp = memory.copy()

        # We couldn't have only 1 layer on first or last GPU, so we remove it to avoid 
        #   this case.
        repartition[0]  -= 1
        repartition[-1] -= 1
        
        n = len(layers)

        lower_idx = min(range(n), key=lambda i: memory_tmp[i])

        while True:
            upper_idx = max(range(n), key=lambda i: memory_tmp[i])
            memory_tmp[upper_idx] = 0
            if repartition[upper_idx] > 1:
                break
        
        repartition[lower_idx] += 1
        repartition[upper_idx] -= 1
        
        # We restablish the two deleted layers.
        repartition[0]  += 1
        repartition[-1] += 1


        return repartition

    def reset_repartition(self, layer_per_gpu):
        """Reset the distribution of layers based on the number of layers per GPU.
    
        :param layer_per_gpu: Number of layers per GPU.
        :type layer_per_gpu: list
        """
        current_layer = 0
        gpu_index = 0
        separation_layer_index = layer_per_gpu[gpu_index] - 1

        for _, layer in self.LayerLists.items():
            if current_layer >= len(self.LayerLists.items()) - 1:
                break
            
            if separation_layer_index == current_layer:
                layer.reset_separation_layer()
                gpu_index += 1
                separation_layer_index += layer_per_gpu[gpu_index]

            current_layer += 1

    def set_repartition(self, layer_per_gpu):
        """Set the distribution of layers across GPUs based on the number of layers per GPU.
    
        :param layer_per_gpu: Number of layers per GPU.
        :type layer_per_gpu: list
        """
        current_layer = 0
        gpu_index = 0
        separation_layer_index = layer_per_gpu[gpu_index] - 1

        for _, layer in self.LayerLists.items():

            if current_layer >= len(self.LayerLists.items()) - 1:
                self.file += layer.get_declaration()
                break
            
            if separation_layer_index == current_layer:
                layer.set_separation_layer()
                gpu_index += 1
                separation_layer_index += layer_per_gpu[gpu_index]

            self.file += layer.get_declaration()
            current_layer += 1

        self._generate_end_class()

    def _check_memory_peak(self, memory_peak):
        """Check if the memory peaks are balanced across GPUs.
    
        :param memory_peak: Memory peaks for each GPU.
        :type memory_peak: list
        :return: True if memory peaks are balanced, False otherwise.
        :rtype: bool
        """
        threshold = 0.2 
        reference_value = memory_peak[0]
        return all(abs(value - reference_value) <= reference_value * threshold for value in memory_peak[1:])


    def _repartition(self):
        """Perform the distribution of layers across GPUs in a balanced manner based on memory usage."""
        self._init_file()
        # Save self var for remake
        file = self.file
        # Calculate first naive repartition on gpus
        clone_step = len(self.LayerLists.items()) // self.nb_gpu
        remainder = len(self.LayerLists.items()) % self.nb_gpu
        layer_per_gpu = [clone_step] * self.nb_gpu

        # Distribute remainder layer to the GPU
        for i in range(remainder):
            layer_per_gpu[i] += 1

        # Initialise cloned layers
        self.set_repartition(layer_per_gpu)

        # Write in file the naive repartition
        self._write_in_file()

        dir_path = Path(__file__).resolve().parent / "evaluate_mem.py"

        previous_repartitions = []

        while True:
            if layer_per_gpu[0] == 1 or layer_per_gpu[:-1] == 1:
                raise Exception(f"The model is not separable in {self.nb_gpu} GPU, first or last GPU contain only 1 layer. {layer_per_gpu}")

            if 0 in layer_per_gpu:
                raise Exception(f"At least one GPU have 0 layer, please allocate less GPU. {layer_per_gpu}")

            p = subprocess.run([sys.executable, dir_path,
                            '--input_shape', str(list(self.input_shape)),
                            '--output_shape', str(list(self.output_shape)),
                            '--number_gpu', str(int(self.nb_gpu)),
                            '--number_chunks', str(2),
                            '--dtype', str(self.dtype)], capture_output=True, text=True)

            result = p.stdout
            if "CUDA" in result:
                raise Exception(f"The model is too big for the number of GPU given. CUDA OOM")
            
            elif result == '':
                raise Exception(f"An error occured during training. The repartition tried was {layer_per_gpu}.")

            result = result.replace("[", "").replace("]", "")
            result = result.split(",")
            memory_peak = [int(x.strip()) for x in result]
            
            if not self._check_memory_peak(memory_peak):
                new_layer_per_gpu = self._balancing(layer_per_gpu, memory_peak)
                if new_layer_per_gpu in previous_repartitions:
                    break
                
                previous_repartitions.append(new_layer_per_gpu)

                self.file = file
                self.reset_repartition(layer_per_gpu)
                self.set_repartition(new_layer_per_gpu)
                self._write_in_file()
                layer_per_gpu = new_layer_per_gpu

            else:
                break
    
    

    def _filter_trace(self, trace):
        """Filter trace by removing blacklisted layers and un-propaged layers.

        :param trace: Trace of the model
        :type trace: torch.fx.graph._node_list
        """
        to_remove = []        
        filename = Path(__file__).resolve().parent / 'blacklist.txt'
        blacklist = open(filename).readlines()
        blacklist = [line.strip() for line in blacklist]

        for node in trace.graph.nodes:
            if any(item in node.name for item in blacklist):
                to_remove.append(node)

        for node in to_remove[::-1]:
            trace.graph.erase_node(node)

        all_unused_deleted = True
        while all_unused_deleted == True:
            all_unused_deleted = False
            for node in trace.graph.nodes:
                if node.name == "output":
                    break 
                try:
                    trace.graph.erase_node(node)
                    all_unused_deleted = True
                except:
                    continue


    def _tracer(self, net):
        """Trace and create all the composite needed to describe correctly the models given.
        
        It will call the class of class_impl folder to generate at the end the correct file of the model splited between GPUs.
        :param net: Model to trace.
        :type net: a model extended of nn.Module.
        """
        self.net = net

        try:
            trace = symbolic_trace(net)
        except Exception as e:
            raise Exception(f"The model given cannot be traced by torch.fx. Error : {e}")
        
        self._filter_trace(trace)

        self._catch_module_desc()

        prev_node = None

        # Iter through each node traced by torch.fx
        for node in trace.graph.nodes:
            if str(node).find("getitem") >= 0:
                for _node in trace.graph.nodes:
                    if node in _node.args:
                        if str(node.args[0]).find("getattr") >= 0:
                            if self.GetattrLists[node.args[0]].getitem_idx is None:
                                self.GetattrLists[node.args[0]].add_getitem(node.args[1])
                            else:
                                self.GetattrLists[node] = GetAttr(node.args[0], trace.graph.nodes, _node)
                                self.GetattrLists[node].add_getitem(node.args[1])
                            
                        else:
                            self.LayerLists[node.args[0]].add_getitem(node.args[1])
                            
            elif str(node).find("getattr") >= 0:
                self.GetattrLists[node] = GetAttr(node, trace.graph.nodes)

            else:
                if node.op == "call_module":
                    self.LayerLists[node] = CallModule(node, trace.graph.nodes, prev_node,
                                                       self.module_desc[node.target])

                elif node.op == "call_function" or node.op == "call_method":
                    self.LayerLists[node] = CallFunction(node, trace.graph.nodes, prev_node)
                elif node.op == "get_attr":
                    self.LayerLists[node] = GetAttrModule(node, trace.graph.nodes, prev_node, net)
                    pass

                else:
                    self.LayerLists[node] = PropagationLayer(node, trace.graph.nodes, prev_node)
                prev_node = node

        # For each getattr, we will update the Layer who are linked to it. If the value of the getattr need to be
        # stashed we will update the stash list of the parent of the getattr. And by default update all the argument
        # to have the correct declaration with the getattr.
        for _, getattr_item in self.GetattrLists.items():
            # if getattr_item.get_child() is not None:
            if getattr_item.is_stash_needed():
                self.LayerLists[getattr_item.get_parent()].add_stash(getattr_item.get_child())

            self.LayerLists[getattr_item.get_child()].update_arg_by_attr(getattr_item.get_attr_name(), getattr_item.get_position())

        # As it is complicated to trace efficiently the stash we updated it with the poplist of each node.
        # So for each "pop parent" we set a stash for the current node.
        for layer in self.LayerLists.values():
            for pop_parent in layer.get_pop_parent():
                self.LayerLists[pop_parent[0]].set_stash(layer.get_node())


        # Test mode without GPUs
        if self.nb_gpu == 0:
            print("Pipeline tool as performed a full tracing, but is not allowed to use GPU.")
        else:
            self._repartition()